

#include <wchar.h>
#include "Parser.h"
#include "Scanner.h"




void Parser::SynErr(int n) {
	if (errDist >= minErrDist) errors->SynErr(la->line, la->col, n);
	errDist = 0;
}

void Parser::SemErr(const wchar_t* msg) {
	if (errDist >= minErrDist) errors->Error(t->line, t->col, msg);
	errDist = 0;
}

void Parser::Get() {
	for (;;) {
		t = la;
		la = scanner->Scan();
		if (la->kind <= maxT) { ++errDist; break; }

		if (dummyToken != t) {
			dummyToken->kind = t->kind;
			dummyToken->pos = t->pos;
			dummyToken->col = t->col;
			dummyToken->line = t->line;
			dummyToken->next = NULL;
			coco_string_delete(dummyToken->val);
			dummyToken->val = coco_string_create(t->val);
			t = dummyToken;
		}
		la = t;
	}
}

void Parser::Expect(int n) {
	if (la->kind==n) Get(); else { SynErr(n); }
}

void Parser::ExpectWeak(int n, int follow) {
	if (la->kind == n) Get();
	else {
		SynErr(n);
		while (!StartOf(follow)) Get();
	}
}

bool Parser::WeakSeparator(int n, int syFol, int repFol) {
	if (la->kind == n) {Get(); return true;}
	else if (StartOf(repFol)) {return false;}
	else {
		SynErr(n);
		while (!(StartOf(syFol) || StartOf(repFol) || StartOf(0))) {
			Get();
		}
		return StartOf(syFol);
	}
}

void Parser::object() {
		Expect(2 /* "{" */);
		if (StartOf(1)) {
			tag();
			while (la->kind == 3 /* "," */) {
				Get();
				tag();
			}
		}
		Expect(4 /* "}" */);
}

void Parser::tag() {
		if (StartOf(2)) {
			unpaired_tag();
		} else if (StartOf(3)) {
			paired_tag();
		} else if (la->kind == 8 /* "#" */) {
			text_tag();
		} else SynErr(24);
}

void Parser::unpaired_tag() {
		unpaired_list();
		output.write ("<");
		output.write_tag_buffer();
		
		if (la->kind == 5 /* ":" */) {
			Get();
			Expect(6 /* "[" */);
			attributes();
			Expect(7 /* "]" */);
		}
		output.write (">");
		
}

void Parser::paired_tag() {
		paired_list();
		output.write ("<");
		output.write_tag_buffer();
		
		Expect(5 /* ":" */);
		if (la->kind == 6 /* "[" */) {
			Get();
			attributes();
			Expect(7 /* "]" */);
		}
		output.write (">");
		
		object();
		output.write ("</");
		output.write_tag_buffer();
		output.write (">");
		
}

void Parser::text_tag() {
		Expect(8 /* "#" */);
		Expect(5 /* ":" */);
		Expect(_String);
		output.write_without_quotes (t->val); 
}

void Parser::unpaired_list() {
		switch (la->kind) {
		case 17 /* "br" */: {
			Get();
			output.push_unpaired_tag ("br"); 
			break;
		}
		case 18 /* "hr" */: {
			Get();
			output.push_unpaired_tag ("hr"); 
			break;
		}
		case 19 /* "img" */: {
			Get();
			output.push_unpaired_tag ("img"); 
			break;
		}
		case 20 /* "input" */: {
			Get();
			output.push_unpaired_tag ("input"); 
			break;
		}
		case 21 /* "link" */: {
			Get();
			output.push_unpaired_tag ("link"); 
			break;
		}
		case 22 /* "meta" */: {
			Get();
			output.push_unpaired_tag ("meta"); 
			break;
		}
		default: SynErr(25); break;
		}
}

void Parser::attributes() {
		Expect(_String);
		output.write (" ");output.write_without_quotes (t->val); 
		Expect(5 /* ":" */);
		output.write ("="); 
		Expect(_String);
		output.write (t->val); 
		while (la->kind == 3 /* "," */) {
			Get();
			Expect(_String);
			output.write (" ");output.write_without_quotes (t->val); 
			Expect(5 /* ":" */);
			output.write ("="); 
			Expect(_String);
			output.write (t->val); 
		}
}

void Parser::paired_list() {
		switch (la->kind) {
		case 9 /* "a" */: {
			Get();
			output.push_paired_tag ("a"); 
			break;
		}
		case 10 /* "b" */: {
			Get();
			output.push_paired_tag ("b"); 
			break;
		}
		case 11 /* "body" */: {
			Get();
			output.push_paired_tag ("body"); 
			break;
		}
		case 12 /* "form" */: {
			Get();
			output.push_paired_tag ("form"); 
			break;
		}
		case 13 /* "head" */: {
			Get();
			output.push_paired_tag ("head"); 
			break;
		}
		case 14 /* "html" */: {
			Get();
			output.push_paired_tag ("html"); 
			break;
		}
		case 15 /* "p" */: {
			Get();
			output.push_paired_tag ("p"); 
			break;
		}
		case 16 /* "div" */: {
			Get();
			output.push_paired_tag ("div"); 
			break;
		}
		default: SynErr(26); break;
		}
}




// If the user declared a method Init and a mehtod Destroy they should
// be called in the contructur and the destructor respctively.
//
// The following templates are used to recognize if the user declared
// the methods Init and Destroy.

template<typename T>
struct ParserInitExistsRecognizer {
	template<typename U, void (U::*)() = &U::Init>
	struct ExistsIfInitIsDefinedMarker{};

	struct InitIsMissingType {
		char dummy1;
	};
	
	struct InitExistsType {
		char dummy1; char dummy2;
	};

	// exists always
	template<typename U>
	static InitIsMissingType is_here(...);

	// exist only if ExistsIfInitIsDefinedMarker is defined
	template<typename U>
	static InitExistsType is_here(ExistsIfInitIsDefinedMarker<U>*);

	enum { InitExists = (sizeof(is_here<T>(NULL)) == sizeof(InitExistsType)) };
};

template<typename T>
struct ParserDestroyExistsRecognizer {
	template<typename U, void (U::*)() = &U::Destroy>
	struct ExistsIfDestroyIsDefinedMarker{};

	struct DestroyIsMissingType {
		char dummy1;
	};
	
	struct DestroyExistsType {
		char dummy1; char dummy2;
	};

	// exists always
	template<typename U>
	static DestroyIsMissingType is_here(...);

	// exist only if ExistsIfDestroyIsDefinedMarker is defined
	template<typename U>
	static DestroyExistsType is_here(ExistsIfDestroyIsDefinedMarker<U>*);

	enum { DestroyExists = (sizeof(is_here<T>(NULL)) == sizeof(DestroyExistsType)) };
};

// The folloing templates are used to call the Init and Destroy methods if they exist.

// Generic case of the ParserInitCaller, gets used if the Init method is missing
template<typename T, bool = ParserInitExistsRecognizer<T>::InitExists>
struct ParserInitCaller {
	static void CallInit(T *t) {
		// nothing to do
	}
};

// True case of the ParserInitCaller, gets used if the Init method exists
template<typename T>
struct ParserInitCaller<T, true> {
	static void CallInit(T *t) {
		t->Init();
	}
};

// Generic case of the ParserDestroyCaller, gets used if the Destroy method is missing
template<typename T, bool = ParserDestroyExistsRecognizer<T>::DestroyExists>
struct ParserDestroyCaller {
	static void CallDestroy(T *t) {
		// nothing to do
	}
};

// True case of the ParserDestroyCaller, gets used if the Destroy method exists
template<typename T>
struct ParserDestroyCaller<T, true> {
	static void CallDestroy(T *t) {
		t->Destroy();
	}
};

void Parser::Parse() {
	t = NULL;
	la = dummyToken = new Token();
	la->val = coco_string_create(L"Dummy Token");
	Get();
	object();
	Expect(0);
}

Parser::Parser(Scanner *scanner) {
	maxT = 23;

	ParserInitCaller<Parser>::CallInit(this);
	dummyToken = NULL;
	t = la = NULL;
	minErrDist = 2;
	errDist = minErrDist;
	this->scanner = scanner;
	errors = new Errors();
}

bool Parser::StartOf(int s) {
	const bool T = true;
	const bool x = false;

	static bool set[4][25] = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x},
		{x,x,x,x, x,x,x,x, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,x, x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,T,T,T, T,T,T,x, x},
		{x,x,x,x, x,x,x,x, x,T,T,T, T,T,T,T, T,x,x,x, x,x,x,x, x}
	};



	return set[s][la->kind];
}

Parser::~Parser() {
	ParserDestroyCaller<Parser>::CallDestroy(this);
	delete errors;
	delete dummyToken;
}

Errors::Errors() {
	count = 0;
}

void Errors::SynErr(int line, int col, int n) {
	wchar_t* s;
	switch (n) {
			case 0: s = coco_string_create(L"EOF expected"); break;
			case 1: s = coco_string_create(L"String expected"); break;
			case 2: s = coco_string_create(L"\"{\" expected"); break;
			case 3: s = coco_string_create(L"\",\" expected"); break;
			case 4: s = coco_string_create(L"\"}\" expected"); break;
			case 5: s = coco_string_create(L"\":\" expected"); break;
			case 6: s = coco_string_create(L"\"[\" expected"); break;
			case 7: s = coco_string_create(L"\"]\" expected"); break;
			case 8: s = coco_string_create(L"\"#\" expected"); break;
			case 9: s = coco_string_create(L"\"a\" expected"); break;
			case 10: s = coco_string_create(L"\"b\" expected"); break;
			case 11: s = coco_string_create(L"\"body\" expected"); break;
			case 12: s = coco_string_create(L"\"form\" expected"); break;
			case 13: s = coco_string_create(L"\"head\" expected"); break;
			case 14: s = coco_string_create(L"\"html\" expected"); break;
			case 15: s = coco_string_create(L"\"p\" expected"); break;
			case 16: s = coco_string_create(L"\"div\" expected"); break;
			case 17: s = coco_string_create(L"\"br\" expected"); break;
			case 18: s = coco_string_create(L"\"hr\" expected"); break;
			case 19: s = coco_string_create(L"\"img\" expected"); break;
			case 20: s = coco_string_create(L"\"input\" expected"); break;
			case 21: s = coco_string_create(L"\"link\" expected"); break;
			case 22: s = coco_string_create(L"\"meta\" expected"); break;
			case 23: s = coco_string_create(L"??? expected"); break;
			case 24: s = coco_string_create(L"invalid tag"); break;
			case 25: s = coco_string_create(L"invalid unpaired_list"); break;
			case 26: s = coco_string_create(L"invalid paired_list"); break;

		default:
		{
			wchar_t format[20];
			coco_swprintf(format, 20, L"error %d", n);
			s = coco_string_create(format);
		}
		break;
	}
	wprintf(L"-- line %d col %d: %ls\n", line, col, s);
	coco_string_delete(s);
	count++;
}

void Errors::Error(int line, int col, const wchar_t *s) {
	wprintf(L"-- line %d col %d: %ls\n", line, col, s);
	count++;
}

void Errors::Warning(int line, int col, const wchar_t *s) {
	wprintf(L"-- line %d col %d: %ls\n", line, col, s);
}

void Errors::Warning(const wchar_t *s) {
	wprintf(L"%ls\n", s);
}

void Errors::Exception(const wchar_t* s) {
	wprintf(L"%ls", s); 
	exit(1);
}


